@@ -1,16 +1,16 @@
 package Module::ScanDeps;
 
-use 5.004;
+use 5.006;
 use strict;
-use vars qw( $VERSION @EXPORT @EXPORT_OK $CurrentPackage @IncludeLibs );
+use vars qw( $VERSION @EXPORT @EXPORT_OK @ISA $CurrentPackage @IncludeLibs $ScanFileRE );
 
-$VERSION   = '0.74';
+$VERSION   = '0.89';
 @EXPORT    = qw( scan_deps scan_deps_runtime );
-@EXPORT_OK = qw( scan_line scan_chunk add_deps scan_deps_runtime );
+@EXPORT_OK = qw( scan_line scan_chunk add_deps scan_deps_runtime path_to_inc_name );
 
 use Config;
-use Exporter;
-use base 'Exporter';
+require Exporter;
+our @ISA = qw(Exporter);
 use constant dl_ext  => ".$Config{dlext}";
 use constant lib_ext => $Config{lib_ext};
 use constant is_insensitive_fs => (
@@ -19,22 +19,21 @@ use constant is_insensitive_fs => (
         and (-s lc($0) || -1) == -s $0
 );
 
+use version;
 use Cwd ();
 use File::Path ();
 use File::Temp ();
 use File::Spec ();
 use File::Basename ();
 use FileHandle;
+use Module::Build::ModuleInfo;
+
+$ScanFileRE = qr/(?:^|\\|\/)(?:[^.]*|.*\.(?i:p[ml]|t|al))$/;
 
 =head1 NAME
 
 Module::ScanDeps - Recursively scan Perl code for dependencies
 
-=head1 VERSION
-
-This document describes version 0.71 of Module::ScanDeps, released
-January  4, 2007.
-
 =head1 SYNOPSIS
 
 Via the command-line program L<scandeps.pl>:
@@ -76,11 +75,12 @@ hash reference; its keys are the module names as appears in C<%INC>
         key     => 'Test/More.pm',
         type    => 'module',    # or 'autoload', 'data', 'shared'
         used_by => [ 'Test/Simple.pm', ... ],
+        uses    => [ 'Test/Other.pm', ... ],
     }
 
-One function, C<scan_deps>, is exported by default.  Three other
-functions (C<scan_line>, C<scan_chunk>, C<add_deps>) are exported upon
-request.
+One function, C<scan_deps>, is exported by default.  Other
+functions such as (C<scan_line>, C<scan_chunk>, C<add_deps>, C<path_to_inc_name>)
+are exported upon request.
 
 Users of B<App::Packer> may also use this module as the dependency-checking
 frontend, by tweaking their F<p2e.pl> like below:
@@ -121,7 +121,7 @@ If C<$execute> is an array reference, runs the files contained
 in it instead of C<@files>.
 
 Additionally, an option C<warn_missing> is recognized. If set to true,
-C<scan_deps> issues a warning to STDOUT for every module file that the
+C<scan_deps> issues a warning to STDERR for every module file that the
 scanned code depends but that wasn't found. Please note that this may
 also report numerous false positives. That is why by default, the heuristic
 silently drops all dependencies it cannot find.
@@ -165,11 +165,34 @@ modules that cannot be found are skipped.
 This function populates the C<%rv> hash with module/filename pairs, and
 returns a reference to it.
 
+=head2 B<path_to_inc_name>
+
+    $perl_name = path_to_inc_name($path, $warn)
+
+Assumes C<$path> refers to a perl file and does it's best to return the
+name as it would appear in %INC. Returns undef if no match was found 
+and a prints a warning to STDERR if C<$warn> is true.
+
+E.g. if C<$path> = perl/site/lib/Module/ScanDeps.pm then C<$perl_name>
+will be Module/ScanDeps.pm.
+
 =head1 NOTES
 
-You can set the global variable C<@Module::ScanDeps::IncludeLibs> to
-specify additional directories in which to search modules
-without modifying C<@INC> itself.
+=head2 B<@Module::ScanDeps::IncludeLibs>
+
+You can set this global variable to specify additional directories in
+which to search modules without modifying C<@INC> itself.
+
+=head2 B<$Module::ScanDeps::ScanFileRE>
+
+You can set this global variable to specify a regular expression to 
+identify what files to scan. By default it includes all files of 
+the following types: .pm, .pl, .t and .al. Additionally, all files
+without a suffix are considered.
+
+For instance, if you want to scan all files then use the following:
+
+C<$Module::ScanDeps::ScanFileRE = qr/./>
 
 =head1 CAVEATS
 
@@ -228,6 +251,7 @@ my %Preload;
     },
     'Catalyst/Engine.pm' => 'sub',
     'Class/MakeMethods.pm' => 'sub',
+    'Class/MethodMaker.pm' => 'sub',
     'Config/Any.pm' =>'sub',
     'Crypt/Random.pm' => sub {
         _glob_in_inc('Crypt/Random/Provider', 1);
@@ -250,6 +274,7 @@ my %Preload;
         grep /\bMM_/, _glob_in_inc('ExtUtils', 1);
     },
     'File/Basename.pm' => [qw( re.pm )],
+    'File/HomeDir.pm' => 'sub',
     'File/Spec.pm'     => sub {
         require File::Spec;
         map { my $name = $_; $name =~ s!::!/!g; "$name.pm" } @File::Spec::ISA;
@@ -257,6 +282,11 @@ my %Preload;
     'HTTP/Message.pm' => [ qw(
         URI/URL.pm          URI.pm
     ) ],
+    'Image/ExifTool.pm' => sub {
+        return( _glob_in_inc("Image/ExifTool", 1), qw(
+            File/RandomAccess.pm
+        ));
+    },
     'Image/Info.pm' => sub {
         return( _glob_in_inc("Image/Info", 1), qw(
             Image/TIFF.pm
@@ -301,21 +331,23 @@ my %Preload;
     'MIME/Decoder.pm'               => 'sub',
     'Net/DNS/RR.pm'                 => 'sub',
     'Net/FTP.pm'                    => 'sub',
+    'Net/Server.pm'                 => 'sub',
     'Net/SSH/Perl.pm'               => 'sub',
     'PDF/API2/Resource/Font.pm'     => 'sub',
     'PDF/API2/Basic/TTF/Font.pm'    => sub {
         _glob_in_inc('PDF/API2/Basic/TTF', 1);
     },
     'PDF/Writer.pm'                 => 'sub',
-    'POE'                           => [ qw(
+    'POE.pm'                           => [ qw(
         POE/Kernel.pm POE/Session.pm
     ) ],
-    'POE/Kernel.pm'                    => [
-        map "POE/Resource/$_.pm", qw(
-            Aliases Events Extrefs FileHandles
-            SIDs Sessions Signals Statistics
-        )
-    ],
+    'POE/Kernel.pm'                    => sub {
+        _glob_in_inc('POE/XS/Resource', 1),
+        _glob_in_inc('POE/Resource', 1),
+        _glob_in_inc('POE/XS/Loop', 1),
+        _glob_in_inc('POE/Loop', 1),
+    },
+    'PPI.pm'                        => 'sub',
     'Parse/AFP.pm'                  => 'sub',
     'Parse/Binary.pm'               => 'sub',
     'PerlIO.pm'                     => [ 'PerlIO/scalar.pm' ],
@@ -341,7 +373,7 @@ my %Preload;
     },
     'Template.pm'      => 'sub',
     'Term/ReadLine.pm' => 'sub',
-	'Test/Deep.pm'     => 'sub',
+    'Test/Deep.pm'     => 'sub',
     'Tk.pm'            => sub {
         $SeenTk = 1;
         qw( Tk/FileSelect.pm Encode/Unicode.pm );
@@ -405,6 +437,9 @@ my %Preload;
 
         return 'pod/perldiag.pod';
     },
+    'threads/shared.pm' => [qw( attributes.pm )],
+    # anybody using threads::shared is likely to declare variables
+    # with attribute :shared
     'utf8.pm' => [
         'utf8_heavy.pl', do {
             my $dir = 'unicore';
@@ -438,10 +473,28 @@ my %Preload;
 
 # }}}
 
-sub _path_to_filename {
-    my $file = shift @_;
-    my ($vol, $dir, $key) = File::Spec->splitpath($file);
-    return $key;
+sub path_to_inc_name($$) {
+    my $path = shift;
+    my $warn = shift;
+    my $inc_name;
+
+    if ($path =~ m/\.pm$/io) {
+        die "$path doesn't exist" unless (-f $path);
+        my $module_info = Module::Build::ModuleInfo->new_from_file($path);
+        die "Module::Build::ModuleInfo error: $!" unless defined($module_info);
+        $inc_name = $module_info->name();
+        if (defined($inc_name)) {
+            $inc_name =~ s|\:\:|\/|og;
+            $inc_name .= '.pm';
+        } else {
+            warn "# Couldn't find include name for $path\n" if $warn;
+        }
+    } else {
+        # Bad solution!
+        (my $vol, my $dir, $inc_name) = File::Spec->splitpath($path);
+    }
+
+    return $inc_name;
 }
 
 my $Keys = 'files|keys|recurse|rv|skip|first|execute|compile|warn_missing';
@@ -451,16 +504,39 @@ sub scan_deps {
         (@_ and $_[0] =~ /^(?:$Keys)$/o) ? @_ : (files => [@_], recurse => 1)
     );
 
-    if (!defined($args{keys})) { 
-        $args{keys} = [map {_path_to_filename($_)} @{$args{files}}]
+    if (!defined($args{keys})) {
+        $args{keys} = [map {path_to_inc_name($_, $args{warn_missing})} @{$args{files}}];
     }
 
     my ($type, $path);
     foreach my $input_file (@{$args{files}}) {
+        if ($input_file !~ $ScanFileRE) {
+            warn "Skipping input file $input_file because it matches \$Module::ScanDeps::ScanFileRE\n" if $args{warn_missing};
+            next;
+        }
+
         $type = 'module';
         $type = 'data' unless $input_file =~ /\.p[mh]$/io;
         $path = $input_file;
-        _add_info($args{rv}, _path_to_filename($path), $path, undef, $type);
+        if ($type eq 'module') {
+            # necessary because add_deps does the search for shared libraries and such
+            add_deps(
+                used_by => undef,
+                rv => $args{rv},
+                modules => [path_to_inc_name($path, $args{warn_missing})],
+                skip => undef,
+                warn_missing => $args{warn_missing},
+            );
+        }
+        else {
+            _add_info(
+                rv      => $args{rv},
+                module  => path_to_inc_name($path, $args{warn_missing}),
+                file    => $path,
+                used_by => undef,
+                type    => $type,
+            );
+        }
     }
 
     scan_deps_static(\%args);
@@ -475,22 +551,26 @@ sub scan_deps {
         );
     }
 
+    # do not include the input files themselves as dependencies!
+    delete $args{rv}{$_} foreach @{$args{files}};
+
     return ($args{rv});
 }
 
 sub scan_deps_static {
     my ($args) = @_;
-    my ($files, $keys, $recurse, $rv, $skip, $first, $execute, $compile) =
-        @$args{qw( files keys recurse rv skip first execute compile )};
+    my ($files, $keys, $recurse, $rv, $skip, $first, $execute, $compile, $_skip) =
+        @$args{qw( files keys recurse rv skip first execute compile _skip )};
 
     $rv   ||= {};
-    $skip ||= {};
+    $_skip ||= { %{$skip || {}} };
 
     foreach my $file (@{$files}) {
         my $key = shift @{$keys};
-        next if $skip->{$file}++;
+        next if $_skip->{$file}++;
         next if is_insensitive_fs()
-          and $file ne lc($file) and $skip->{lc($file)}++;
+          and $file ne lc($file) and $_skip->{lc($file)}++;
+        next unless $file =~ $ScanFileRE;
 
         local *FH;
         open FH, $file or die "Cannot open $file: $!";
@@ -502,14 +582,12 @@ sub scan_deps_static {
             chomp(my $line = $_);
             foreach my $pm (scan_line($line)) {
                 last LINE if $pm eq '__END__';
-                
-                # Skip Tk hits from Term::ReadLine
+
+                # Skip Tk hits from Term::ReadLine and Tcl::Tk
                 my $pathsep = qr/\/|\\|::/;
-                if (
-                    $file =~ /${pathsep}Term${pathsep}ReadLine.pm$/
-                    and $pm =~ /^Tk\b/
-                ) {
-                    next;
+                if ($pm =~ /^Tk\b/) {
+                  next if $file =~ /(?:^|${pathsep})Term${pathsep}ReadLine\.pm$/;
+                  next if $file =~ /(?:^|${pathsep})Tcl${pathsep}Tk\W/;
                 }
 
                 if ($pm eq '__POD__') {
@@ -517,7 +595,7 @@ sub scan_deps_static {
                     next LINE;
                 }
 
-                $pm = 'CGI/Apache.pm' if /^Apache(?:\.pm)$/;
+                $pm = 'CGI/Apache.pm' if $file =~ /^Apache(?:\.pm)$/;
 
                 add_deps(
                     used_by => $key,
@@ -553,6 +631,7 @@ sub scan_deps_static {
             rv      => $rv,
             skip    => $skip,
             recurse => 0,
+            _skip   => $_skip,
         }) or ($args->{_deep} and return);
         last if $count == keys %$rv;
     }
@@ -578,6 +657,8 @@ sub scan_deps_runtime {
         my $file;
 
         foreach $file (@$files) {
+            next unless $file =~ $ScanFileRE;
+
             ($inchash, $dl_shared_objects, $incarray) = ({}, [], []);
             _compile($perl, $file, $inchash, $dl_shared_objects, $incarray);
 
@@ -623,8 +704,15 @@ sub scan_line {
             $CurrentPackage =~ s{::}{/}g;
             return;
         }
-        return if /^\s*(use|require)\s+[\d\._]+/;
-        if (my ($autouse) = /^\s*use\s+autouse\s+(["'].*?["']|\w+)/)	
+        # use VERSION:
+        if (/^\s*(?:use|require)\s+([\d\._]+)/) {
+          # include feaure.pm if we have 5.9.5 or better
+          if (version->new($1) >= version->new("5.9.5")) { # seems to catch 5.9, too (but not 5.9.4)
+            return "feature.pm";
+          }
+        }
+
+        if (my ($autouse) = /^\s*use\s+autouse\s+(["'].*?["']|\w+)/)
         {
             $autouse =~ s/["']//g;
             $autouse =~ s{::}{/}g;
@@ -649,6 +737,25 @@ sub scan_line {
     return sort keys %found;
 }
 
+# short helper for scan_chunk
+sub _typical_module_loader_chunk {
+  local $_ = shift;
+  my $loader = shift;
+  my $loader_file = $loader;
+  $loader_file =~ s/::/\//;
+  $loader_file .= ".pm";
+  $loader = quotemeta($loader);
+
+  if (/^\s* use \s+ $loader \b \s* (.*)/sx) {
+    return [
+      $loader_file,
+      map { s{::}{/}g; "$_.pm" }
+      grep { length and !/^q[qw]?$/ } split(/[^\w:]+/, $1)
+    ];
+  }
+  return();
+}
+
 sub scan_chunk {
     my $chunk = shift;
 
@@ -656,27 +763,19 @@ sub scan_chunk {
     my $module = eval {
         $_ = $chunk;
 
-        return [ 'base.pm',
-            map { s{::}{/}g; "$_.pm" }
-              grep { length and !/^q[qw]?$/ } split(/[^\w:]+/, $1) ]
-          if /^\s* use \s+ base \s+ (.*)/sx;
+        # TODO: There's many more of these "loader" type modules on CPAN!
+        # scan for the typical module-loader modules
+        foreach my $loader (qw(asa base prefork POE encoding maybe only::matching)) {
+          my $retval = _typical_module_loader_chunk($_, $loader);
+          return $retval if $retval;
+        }
 
         return [ 'Class/Autouse.pm',
             map { s{::}{/}g; "$_.pm" }
               grep { length and !/^:|^q[qw]?$/ } split(/[^\w:]+/, $1) ]
-          if /^\s* use \s+ Class::Autouse \s+ (.*)/sx
+          if /^\s* use \s+ Class::Autouse \b \s* (.*)/sx
               or /^\s* Class::Autouse \s* -> \s* autouse \s* (.*)/sx;
 
-        return [ 'POE.pm',
-            map { s{::}{/}g; "POE/$_.pm" }
-              grep { length and !/^q[qw]?$/ } split(/[^\w:]+/, $1) ]
-          if /^\s* use \s+ POE \s+ (.*)/sx;
-
-        return [ 'encoding.pm',
-            map { _find_encoding($_) }
-              grep { length and !/^q[qw]?$/ } split(/[^\w:-]+/, $1) ]
-          if /^\s* use \s+ encoding \s+ (.*)/sx;
-
         return $1 if /(?:^|\s)(?:use|no|require)\s+([\w:\.\-\\\/\"\']+)/;
         return $1
           if /(?:^|\s)(?:use|no|require)\s+\(\s*([\w:\.\-\\\/\"\']+)\s*\)/;
@@ -699,8 +798,8 @@ sub scan_chunk {
         }
         return $1 if /(?:^|\s)(?:do|require)\s+[^"]*"(.*?)"/;
         return $1 if /(?:^|\s)(?:do|require)\s+[^']*'(.*?)'/;
-        return $1 if /[^\$]\b([\w:]+)->\w/ and $1 ne 'Tk';
-        return $1 if /\b(\w[\w:]*)::\w+\(/;
+        return $1 if /[^\$]\b([\w:]+)->\w/ and $1 ne 'Tk' and $1 ne 'shift';
+        return $1 if /\b(\w[\w:]*)::\w+\(/ and $1 ne 'main' and $1 ne 'SUPER';
 
         if ($SeenTk) {
             my @modules;
@@ -742,26 +841,57 @@ sub _find_encoding {
 }
 
 sub _add_info {
-    my ($rv, $module, $file, $used_by, $type) = @_;
+    my %args = @_;
+    my ($rv, $module, $file, $used_by, $type) = @args{qw/rv module file used_by type/};
+
     return unless defined($module) and defined($file);
 
     # Ensure file is always absolute
     $file = File::Spec->rel2abs($file);
     $file =~ s|\\|\/|go;
 
+    # Avoid duplicates that can arise due to case differences that don't actually 
+    # matter on a case tolerant system
+    if (File::Spec->case_tolerant()) {
+        foreach my $key (keys %$rv) {
+            if (lc($key) eq lc($module)) {
+                $module = $key;
+                last;
+            }
+        }
+        if (defined($used_by)) {
+            if (lc($used_by) eq lc($module)) {
+                $used_by = $module;
+            } else {
+                foreach my $key (keys %$rv) {
+                    if (lc($key) eq lc($used_by)) {
+                        $used_by = $key;
+                        last;
+                    }
+                }
+            }
+        }
+    }
+
     $rv->{$module} ||= {
         file => $file,
         key  => $module,
         type => $type,
     };
 
-    push @{ $rv->{$module}{used_by} }, $used_by
-      if defined($used_by)
-      and $used_by ne $module
-      and !grep { $_ eq $used_by } @{ $rv->{$module}{used_by} };
+    if (defined($used_by) and $used_by ne $module) {
+        push @{ $rv->{$module}{used_by} }, $used_by
+          if  ( (!File::Spec->case_tolerant() && !grep { $_ eq $used_by } @{ $rv->{$module}{used_by} })
+             or ( File::Spec->case_tolerant() && !grep { lc($_) eq lc($used_by) } @{ $rv->{$module}{used_by} }));
+
+        # We assume here that another _add_info will be called to provide the other parts of $rv->{$used_by}    
+        push @{ $rv->{$used_by}{uses} }, $module
+          if  ( (!File::Spec->case_tolerant() && !grep { $_ eq $module } @{ $rv->{$used_by}{uses} })
+             or ( File::Spec->case_tolerant() && !grep { lc($_) eq lc($module) } @{ $rv->{$used_by}{uses} }));
+    }
 }
 
-# This subroutine relies on not being called for modules that should be skipped
+# This subroutine relies on not being called for modules that have already been visited
 sub add_deps {
     my %args =
       ((@_ and $_[0] =~ /^(?:modules|rv|used_by|warn_missing)$/)
@@ -769,20 +899,26 @@ sub add_deps {
         : (rv => (ref($_[0]) ? shift(@_) : undef), modules => [@_]));
 
     my $rv = $args{rv}   || {};
+    my $skip = $args{skip} || {};
     my $used_by = $args{used_by};
 
     foreach my $module (@{ $args{modules} }) {
         my $file = _find_in_inc($module)
           or _warn_of_missing_module($module, $args{warn_missing}), next;
+        next if $skip->{$file};
 
         if (exists $rv->{$module}) {
-            _add_info($rv, $module, $file, $used_by, undef);
+            _add_info( rv     => $rv,      module  => $module,
+                       file   => $file,    used_by => $used_by,
+                       type   => undef );
             next;
         }
 
         my $type = 'module';
         $type = 'data' unless $file =~ /\.p[mh]$/i;
-        _add_info($rv, $module, $file, $used_by, $type);
+        _add_info( rv     => $rv,   module  => $module,
+                   file   => $file, used_by => $used_by,
+                   type   => $type );
 
         if ($module =~ /(.*?([^\/]*))\.p[mh]$/i) {
             my ($path, $basename) = ($1, $2);
@@ -796,24 +932,25 @@ sub add_deps {
                 $type = 'autoload' if $ext eq '.ix' or $ext eq '.al';
                 $type ||= 'data';
 
-                _add_info($rv, "auto/$path/$_->{name}", $_->{file}, $module,
-                    $type);
+                _add_info( rv     => $rv,        module  => "auto/$path/$_->{name}",
+                           file   => $_->{file}, used_by => $module,
+                           type   => $type );
             }
         }
-    }
-
+    } # end for modules
     return $rv;
 }
 
 sub _find_in_inc {
     my $file = shift;
 
-    # absolute file names
-    return $file if -f $file;
-
     foreach my $dir (grep !/\bBSDPAN\b/, @INC, @IncludeLibs) {
         return "$dir/$file" if -f "$dir/$file";
     }
+
+    # absolute file names
+    return $file if -f $file;
+
     return;
 }
 
@@ -1102,7 +1239,12 @@ sub _merge_rv {
 
 sub _not_dup {
     my ($key, $rv1, $rv2) = @_;
-    (_abs_path($rv1->{$key}{file}) ne _abs_path($rv2->{$key}{file}));
+    if (File::Spec->case_tolerant()) {
+        return lc(_abs_path($rv1->{$key}{file})) ne lc(_abs_path($rv2->{$key}{file}));
+    }
+    else {
+        return _abs_path($rv1->{$key}{file}) ne _abs_path($rv2->{$key}{file});
+    }
 }
 
 sub _abs_path {
@@ -1119,7 +1261,7 @@ sub _warn_of_missing_module {
     my $warn = shift;
     return if not $warn;
     return if not $module =~ /\.p[ml]$/;
-    print "# Could not find source file '$module' in \@INC or \@IncludeLibs. Skipping it.\n"
+    warn "# Could not find source file '$module' in \@INC or \@IncludeLibs. Skipping it.\n"
       if not -f $module;
 }
 
@@ -1151,7 +1293,7 @@ documentations on CPAN for further information.
 
 =head1 AUTHORS
 
-Audrey Tang E<lt>autrijus@autrijus.orgE<gt>
+Audrey Tang E<lt>cpan@audreyt.orgE<gt>
 
 To a lesser degree: Steffen Mueller E<lt>smueller@cpan.orgE<gt>
 
@@ -1179,9 +1321,9 @@ Please submit bug reports to E<lt>bug-Module-ScanDeps@rt.cpan.orgE<gt>.
 
 =head1 COPYRIGHT
 
-Copyright 2002, 2003, 2004, 2005, 2006 by
-Audrey Tang E<lt>autrijus@autrijus.orgE<gt>;
-2005, 2006 by Steffen Mueller E<lt>smueller@cpan.orgE<gt>.
+Copyright 2002-2008 by
+Audrey Tang E<lt>cpan@audreyt.orgE<gt>;
+2005-2008 by Steffen Mueller E<lt>smueller@cpan.orgE<gt>.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
